import abc


class Executable(abc.ABC):
    @abc.abstractmethod
    def run(self, data):
        pass


class Serializable(abc.ABC):
    @classmethod
    @abc.abstractclassmethod
    def from_config(cls, factory, config):
        pass

    @abc.abstractmethod
    def to_config(self):
        pass


class Nesteable(abc.ABC):
    @abc.abstractmethod
    def set_parent(self, parent):
        pass

    @abc.abstractmethod
    def get_parent(self):
        pass

    @abc.abstractmethod
    def add_child(self, child):
        pass

    @abc.abstractmethod
    def get_children(self):
        pass

    @abc.abstractmethod
    def remove_child(self, child):
        pass


class AbstractTask(Executable, Serializable, Nesteable):
    @abc.abstractmethod
    def get_id(self):
        pass

    @abc.abstractmethod
    def get_name(self):
        pass



import abc
import json

try:
    import yaml
    yaml_exists = True
except ImportError:
    yaml_exists = False


class Configuration(abc.ABC):
    @abc.abstractmethod
    def get_config_dictionary(self):
        pass


class DictionaryConfiguration(Configuration):
    def __init__(self, config):
        self.config = config

    def get_config_dictionary(self):
        return self.config


class TextSourceStrategy(abc.ABC):
    @abc.abstractmethod
    def get_text(self):
        pass


class StringTextSourceStrategy(TextSourceStrategy):
    def __init__(self, string):
        self.string = string

    def get_text(self):
        return self.string


class FileTextSourceStrategy(TextSourceStrategy):
    def __init__(self, filename):
        self.filename = filename

    def get_text(self):
        with open(self.filename) as file:
            return file.read()


class TextFormatStrategy(abc.ABC):
    @abc.abstractmethod
    def to_dict(self, text):
        pass


class JSONTextFormatStrategy(TextFormatStrategy):
    def to_dict(self, text):
        return json.loads(text)


class YAMLTextFormatStrategy(TextFormatStrategy):
    def to_dict(self, text):
        self.check_pyyaml_installed()
        return yaml.safe_load(text)

    def check_pyyaml_installed(self):
        if yaml_exists is None:
            raise ImportError("Install pyyaml package to use YAML config files")


class TextConfiguration(Configuration):
    def __init__(self, source_strategy, format_strategy):
        self.source_strategy = source_strategy
        self.format_strategy = format_strategy

    def get_config_dictionary(self):
        text = self.source_strategy.get_text()
        data = self.format_strategy.to_dict(text)
        return data


# TODO: implement SQL database configuration reader.
class SQLDatabaseConfiguration(Configuration):
    pass


FORMATS = {
    "json": JSONTextFormatStrategy,
    "yaml": YAMLTextFormatStrategy,
}

SOURCES = {
    "string": StringTextSourceStrategy,
    "file": FileTextSourceStrategy,
}


def from_text(text_or_filename, origin="file", format="json"):
    if type(origin) == str:
        origin = SOURCES[origin](text_or_filename)
    if type(format) == str:
        format = FORMATS[format]()
    conf = TextConfiguration(origin, format).get_config_dictionary()
    return conf



from sisifo.utils.named_collection import NamedCollectionBuilder
from sisifo.utils.named_collection import NamedCollectionDecorator


class DataCollection(NamedCollectionDecorator):
    def __init__(self):
        data = (
            NamedCollectionBuilder()
            .suggestions()
            .build()
        )
        super().__init__(data)



from concurrent.futures import ThreadPoolExecutor

import sisifo


@sisifo.add_task("common")
class ParallelPipeline(sisifo.Task):
    def __init__(self, steps=[], max_workers=None, **kwargs):
        super().__init__(**kwargs)
        self.max_workers = max_workers

        for i in steps:
            self.create_subtask(i)

    def run(self, data):
        with ThreadPoolExecutor(max_workers=self.max_workers) as ex:
            for task in self.get_children():
                ex.submit(task.run, data)



import uuid

from sisifo.abstract_task import AbstractTask
from sisifo.task_factory import TaskFactory
from sisifo.utils import params


class Task(AbstractTask):
    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def __init__(self, id=None, name=None, factory=None):
        self.id = id or str(uuid.uuid4())
        self.name = name or self.__class__.__name__
        self._factory = factory or TaskFactory()
        self._parent = None
        self._children = []

    def to_config(self):
        properties = vars(self)
        public_properties = {
            k: v for k, v in properties.items() if not k.startswith("_")
        }
        return public_properties

    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def run(self, data):
        return data

    def set_parent(self, parent: AbstractTask):
        if not isinstance(parent, AbstractTask):
            raise ValueError("`parent` should be instance of AbstractTask")
        self._parent = parent

    def get_parent(self):
        return self._parent

    def add_child(self, task) -> None:
        if not isinstance(task, AbstractTask):
            raise ValueError("`task` should be instance of AbstractTask")
        self._children.append(task)

    def get_children(self) -> list:
        return self._children

    def remove_child(self, child) -> None:
        self._children.remove(child)

    def create_subtask(self, task_or_config):
        task = task_or_config
        if not isinstance(task_or_config, AbstractTask):
            if type(task_or_config) == dict:
                task = self._factory.create_task(task_or_config)
            else:
                raise ValueError("Cannot infer task from type: {type(task_or_config)}")
        self.add_child(task)
        task.set_parent(self)
        return task


class TaskDecorator(AbstractTask):
    @classmethod
    def from_config(cls, _):
        raise NotImplementedError()

    def __init__(self, task):
        self.task = task

    def to_config(self):
        return self.task.to_config()

    def get_id(self):
        return self.task.get_id()

    def get_name(self):
        return self.task.get_name()

    def run(self, data):
        return self.task.run(data)

    def set_parent(self, parent):
        return self.task.set_parent(parent)

    def get_parent(self):
        return self.task.get_parent()

    def add_child(self, child):
        return self.task.add_child(child)

    def get_children(self):
        return self.task.get_children()

    def remove_child(self, child):
        return self.task.remove_children(child)


class EntityTask(Task):
    def __init__(self, entity=None, entity_in=None, entity_out=None, **kwargs):
        super().__init__(**kwargs)
        self.entity_in, self.entity_out = params.validate_entity_in_out(
                entity, entity_in, entity_out)

    def run(self, data):
        df = data[self.entity_in]
        data[self.entity_out] = self.run_entity(df)

    def run_entity(self, df):
        return df


class EntityColumnTask(EntityTask):
    def __init__(self, column=None, column_in=None, column_out=None, **kwargs):
        super().__init__(**kwargs)
        self.column_in, self.column_out = params.validate_column_in_out(
            column, column_in, column_out)



def validate_column_out(column=None, column_out=None):
    column_out = column_out or column

    if not column_out:
        raise ValueError("Output column must have a valid value")

    return column_out


def validate_column_in_out(column=None, column_in=None, column_out=None):
    column_in = column_in or column
    column_out = column_out or column_in

    if not column_in:
        raise ValueError("Input column must have a valid value")
    if not column_out:
        raise ValueError("Output column must have a valid value")

    return column_in, column_out


def validate_entity_out(entity=None, entity_out=None):
    entity_out = entity_out or entity

    if not entity_out:
        raise ValueError("Output entity must have a valid value")

    return entity_out


def validate_entity_in(entity=None, entity_in=None):
    entity_in = entity_in or entity

    if not entity_in:
        raise ValueError("Input entity must have a valid value")

    return entity_in


def validate_entity_in_out(entity=None, entity_in=None, entity_out=None):
    entity_in = entity_in or entity
    entity_out = entity_out or entity_in

    if not entity_in:
        raise ValueError("Input entity must have a valid value")
    if not entity_out:
        raise ValueError("Output entity must have a valid value")

    return entity_in, entity_out


def validate_list_of_columns(columns):
    if columns is None or type(columns) != list or len(columns) == 0:
        raise ValueError("Columns should be a non-empty list")

    return columns



"""Sísifo task runner.

Generic framework for running tasks over a data collection.
"""

from sisifo._version import __version__  # noqa
from sisifo.abstract_task import AbstractTask  # noqa
from sisifo.concrete_tasks import EntityColumnTask  # noqa
from sisifo.concrete_tasks import EntityTask  # noqa
from sisifo.concrete_tasks import Task  # noqa
from sisifo.concrete_tasks import TaskDecorator  # noqa
from sisifo.data_collection import DataCollection  # noqa
from sisifo.task_factory import TaskFactory  # noqa
from sisifo.task_factory import create_task  # noqa
from sisifo.task_registry import TaskRegistry  # noqa
from sisifo.task_registry import add_task  # noqa
from sisifo.task_registry import get_task  # noqa
from sisifo.task_registry import get_task_registry  # noqa
from sisifo.task_registry import has_task  # noqa
from sisifo.task_registry import remove_task  # noqa
from sisifo.task_registry import replace_task  # noqa
from sisifo.task_registry import set_task_registry  # noqa
from sisifo.utils.named_collection import NamedCollection  # noqa
from sisifo.utils.named_collection import NamedCollectionDecorator  # noqa
from sisifo import namespaces


__author__ = "guiferviz"


def greet():
    """Print a silly sentence. """

    print("The easiest task is the one performed by someone else.\n"
          " - Sísifo Python library")



import sisifo


@sisifo.add_task("common")
class Pipeline(sisifo.Task):
    def __init__(self, steps=[], **kwargs):
        super().__init__(**kwargs)

        for i in steps:
            self.create_subtask(i)

    def run(self, data):
        for task in self.get_children():
            task.run(data)



from sisifo.task_registry import get_task_registry


class TaskFactory:
    def get_task_class_from_name(self, task_name):
        return get_task_registry().get_task(task_name)

    def get_task_class_from_config(self, task_config):
        task_name = task_config.pop("task", None)
        if task_name is None:
            raise ValueError("Task name should be provided")
        task_class = self.get_task_class_from_name(task_name)
        return task_class

    def create_task_from_class(self, task_class, task_factory, task_config):
        task_config["factory"] = task_factory
        return task_class.from_config(task_config)

    def create_task_from_config(self, task_factory, task_config):
        task_class = self.get_task_class_from_config(task_config)
        task = self.create_task_from_class(task_class, task_factory, task_config)
        return task

    def create_task(self, task_config: dict):
        # TODO: apply plugins to `self` TaskFactory
        task_factory = self
        return self.create_task_from_config(task_factory, task_config)


def create_task(task_config: dict):
    return TaskFactory().create_task(task_config)



from sisifo.abstract_task import AbstractTask
from sisifo.utils.named_collection import AllowUpdateDecorator
from sisifo.utils.named_collection import NamedCollectionBuilder


class TaskRegistry:
    def __init__(self):
        self._collection = (
            NamedCollectionBuilder()
            .subclass_of(AbstractTask)
            .suggestions()
            .fail_if_exists()
            .fail_if_name_none()
            .build()
        )

    def add_task(self, canonical_name: str, task: AbstractTask):
        self._collection[canonical_name] = task

    def replace_task(self, canonical_name: str, task: AbstractTask):
        allow_update = AllowUpdateDecorator(self._collection)
        allow_update[canonical_name] = task

    def get_task(self, canonical_name: str):
        return self._collection[canonical_name]

    def remove_task(self, canonical_name: str):
        del self._collection[canonical_name]

    def __contains__(self, canonical_name: str):
        return canonical_name in self._collection

    def get_collection(self):
        return self._collection


_TASK_REGISTRY = TaskRegistry()


def get_task_registry() -> TaskRegistry:
    return _TASK_REGISTRY


def set_task_registry(registry: TaskRegistry) -> None:
    global _TASK_REGISTRY
    _TASK_REGISTRY = registry


def _create_canonical_name(namespace: str, task_name: str) -> str:
    return f"{namespace}.{task_name}"


def _create_canonical_name_from_class(namespace: str, task_class: type) -> str:
    task_name = task_class.__name__
    if namespace:
        return _create_canonical_name(namespace, task_name)
    return task_name


def add_task(namespace: str = ""):
    def wrapper(task_class):
        task_canonical_name = _create_canonical_name_from_class(namespace, task_class)
        get_task_registry().add_task(task_canonical_name, task_class)
        return task_class
    return wrapper


def replace_task(namespace: str = ""):
    def wrapper(task_class):
        task_canonical_name = _create_canonical_name_from_class(namespace, task_class)
        get_task_registry().replace_task(task_canonical_name, task_class)
        return task_class
    return wrapper


def get_task(canonical_name: str) -> type:
    return get_task_registry().get_task(canonical_name)


def remove_task(canonical_name: str) -> None:
    get_task_registry().remove_task(canonical_name)


def has_task(canonical_name: str) -> bool:
    return canonical_name in get_task_registry()



import collections
import difflib


class NamedCollection(collections.UserDict):
    def __missing__(self, key):  # pyrigth needs this because it's not defined in UserDict
        raise KeyError(key)


class BaseNamedCollection(NamedCollection):
    pass


class NamedCollectionDecorator(NamedCollection):
    data: NamedCollection  # pyrigth needs this because data is not an UserDict anymore

    def __init__(self, named_collection: NamedCollection):
        # Do not pass named_collection to super().__init__, it makes a copy.
        super().__init__()
        self.data = named_collection

    def __missing__(self, key):
        return self.data.__missing__(key)


class CheckSubclassDecorator(NamedCollectionDecorator):
    def __init__(self, named_collection, superclass):
        super().__init__(named_collection)
        self.superclass = superclass

    def __setitem__(self, name, element):
        if not issubclass(element, self.superclass):
            superclass_name = self.superclass.__name__
            raise ValueError(f"Expecting class subclass of {superclass_name}")
        super().__setitem__(name, element)


class CheckInstanceDecorator(NamedCollectionDecorator):
    def __init__(self, named_collection, superclass):
        super().__init__(named_collection)
        self.superclass = superclass

    def __setitem__(self, name, element):
        if not isinstance(element, self.superclass):
            superclass_name = self.superclass.__name__
            raise ValueError(f"Expecting instance of {superclass_name}")
        super().__setitem__(name, element)


class KeyErrorWithSuggestions(KeyError):
    def __init__(self, error_key, existing_keys):
        self.error_key = error_key
        self.existing_keys = existing_keys
        super().__init__(self._create_message())

    def _create_message(self):
        message = f"KeyError: '{self.error_key}'"
        matches = self.get_close_matches()
        if len(matches):
            message += f". Similar names: {', '.join(matches)}"
        return message

    def get_close_matches(self, n=3):
        return difflib.get_close_matches(self.error_key, self.existing_keys, n=n)


class SuggestionsDecorator(NamedCollectionDecorator):
    def __missing__(self, name):
        raise KeyErrorWithSuggestions(name, list(self.keys()))


class FailIfNameNoneDecorator(NamedCollectionDecorator):
    def __setitem__(self, name, element):
        if name is None:
            raise ValueError("Name cannot be `None`")
        super().__setitem__(name, element)


class FailIfExistsDecorator(NamedCollectionDecorator):
    def __setitem__(self, name, element):
        if name in self:
            raise ValueError(f"Name '{name}' already in use")
        super().__setitem__(name, element)


class AllowUpdateDecorator(NamedCollectionDecorator):
    def __setitem__(self, name, element):
        if name in self:
            del self[name]
        super().__setitem__(name, element)


class NamedCollectionBuilder:
    def __init__(self):
        self._collection = BaseNamedCollection()

    def subclass_of(self, subclass):
        self._collection = CheckSubclassDecorator(self._collection, subclass)
        return self

    def instance_of(self, subclass):
        self._collection = CheckInstanceDecorator(self._collection, subclass)
        return self

    def suggestions(self):
        self._collection = SuggestionsDecorator(self._collection)
        return self

    def fail_if_name_none(self):
        self._collection = FailIfNameNoneDecorator(self._collection)
        return self

    def fail_if_exists(self):
        self._collection = FailIfExistsDecorator(self._collection)
        return self

    def build(self):
        return self._collection



from sisifo.namespaces import common


def _load_plugins():
    # Import here to avoid polluting "namespaces" module.
    import importlib
    import pkgutil

    for _, full_name, _ in pkgutil.iter_modules():
        if full_name.startswith("sisifo_"):
            _, name = full_name.split("_")
            globals()[name] = importlib.import_module(full_name)


_load_plugins()



import subprocess
import sys

import sisifo


@sisifo.add_task("common")
class RunScript(sisifo.Task):
    def __init__(self, script=None, executable="/bin/bash", **kwargs):
        super().__init__(**kwargs)

        if not script:
            raise ValueError("An script should be provided")

        self.script = script
        self.executable = executable

    def run(self, _):
        subprocess.check_call(self.script, shell=True,
                              stdout=sys.stdout, stderr=sys.stderr,
                              executable=self.executable)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sisifo', 'sisifo.namespaces', 'sisifo.namespaces.common', 'sisifo.utils']

package_data = \
{'': ['*']}

extras_require = \
{'yaml': ['pyyaml>=5.4.1,<6.0.0']}

setup_kwargs = {
    'name': 'sisifo',
    'version': '0.1.0',
    'description': 'Generic framework for running tasks',
    'long_description': "![Sisyphus silhouette](./logo/logo.png)\n\n\n# Sísifo - Task runner\n\nSísifo is the Spanish form of Sisyphus, in ancient Greek: Σίσυφος. This poor\nguy was punished for his self-aggrandizing craftiness and deceitfulness by\nbeing forced to roll an immense boulder up a hill only for it to roll down\nevery time it neared the top, repeating this action for eternity. More\ninformation in [Wikipedia](https://en.wikipedia.org/wiki/Sisyphus).\n\nThis poor library is doomed to an eternity of performing tasks with no other\npurpose in its pitiful and miserable life. I hope you didn't make fun of this\ninsignificant library, our existence is not much more encouraging...\n\n\n# How does it work?\n\nEssentially, Sísifo is just a library that allows you to run tasks on a data\ncollection. Therefore, the most important classes of the library are:\n\n* `sisifo.DataCollection`. See a DataCollection like a dictionary. Use a key to\nstore/retrieve any kind of value from a data collection. The values stored in a\ndata collection are called **entities**.\n* `sisifo.Task`. A task is an action that, usually, modifies the entities in a\ndata collection.\n",
    'author': 'guiferviz',
    'author_email': 'guiferviz@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/guiferviz/sisifo',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
